#!python


__project__ = "alphatims"
__version__ = "1.0.8"
__license__ = "Apache"
__description__ = "A Python package to index Bruker TimsTOF raw data for fast and easy accession and visualization"
__author__ = "Sander Willems, Eugenia Voytik"
__author_email__ = "opensource@alphapept.com"
__github__ = "https://github.com/MannLabs/alphatims"
__keywords__ = [
    "ms",
    "mass spectrometry",
    "bruker",
    "timsTOF",
    "proteomics",
    "bioinformatics",
    "data indexing",
]
__python_version__ = ">=3.8,<4"
__classifiers__ = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]
__console_scripts__ = [
    "alphatims=alphatims.cli:run",
]
__urls__ = {
    "Mann Labs at MPIB": "https://www.biochem.mpg.de/mann",
    "Mann Labs at CPR": "https://www.cpr.ku.dk/research/proteomics/mann/",
    "GitHub": __github__,
    "ReadTheDocs": "https://alphatims.readthedocs.io/en/latest/",
    "PyPi": "https://pypi.org/project/alphatims/",
    "Scientific paper": "https://doi.org/10.1016/j.mcpro.2021.100149",
}
__requirements__ = {
    "": "requirements/requirements.txt",
    "plotting": "requirements/requirements_plotting.txt",
    "development": "requirements/requirements_development.txt",
    "legacy": "requirements/requirements_legacy.txt",
}



#!python -m unittest tests.test_utils
"""This module provides unit tests for alphatims.utils."""

# builtin
import unittest

# local
import alphatims.utils
alphatims.utils.set_progress_callback(None)


class TestLogging(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_set_logger(self):
        pass


class TestStacks(unittest.TestCase):

    def test_option_stack(self):
        value1 = (1, 920)
        value2 = (300, 920)
        value3 = (500, 600)
        scan_stack = alphatims.utils.Option_Stack('scans_slider', value1)
        assert str(scan_stack) == f'0 scans_slider [{value1}]', (
            "The instance of the Option_Stack class is generated wrongly."
        )
        scan_stack.update(value2)
        assert str(scan_stack) == f'1 scans_slider [{value1}, {value2}]', (
            "The update of the class instance with a new value doesn't work."
        )
        scan_stack.update(value2)
        assert str(scan_stack) == f'1 scans_slider [{value1}, {value2}]', (
            "The update of the class instance with the same value doesn't work."
        )
        scan_stack.undo()
        assert str(scan_stack) == f'0 scans_slider [{value1}, {value2}]', (
            "Undo function doesn't work correctly."
        )
        scan_stack.update(value3)
        assert str(scan_stack) == f'1 scans_slider [{value1}, {value3}]', (
            "Update of the values doesn't work after Undo function."
        )
        scan_stack.undo()
        scan_stack.redo()
        assert str(scan_stack) == f'1 scans_slider [{value1}, {value3}]', (
            "Redo function doesn't work correctly."
        )

    def test_global_stack(self):
        value1 = (1, 920)
        value2 = (1, 5000)
        value3 = (300, 920)
        value4 = (1000, 4000)
        value5 = (500, 600)
        value6 = (1000, 2000)
        stack = alphatims.utils.Global_Stack(
            {
                "scans_slider": value1,
                "frames_slider": value2,
            }
        )
        assert str(stack) == f"0 scans_slider [{value1}] 0 frames_slider [{value2}] 0 global [None]", (
            "The instance of the Global_Stack class is generated wrongly."
        )
        stack.update("scans_slider", value3)
        assert str(stack) == f"1 scans_slider [{value1}, {value3}] 0 frames_slider [{value2}] 1 global [None, 'scans_slider']", (
            "The update of the first option inside the class instance with a new value doesn't work."
        )
        stack.update("scans_slider", value3)
        assert str(stack) == f"1 scans_slider [{value1}, {value3}] 0 frames_slider [{value2}] 1 global [None, 'scans_slider']", (
            "The update of the same option with the same value doesn't work."
        )
        stack.update("frames_slider", value4)
        assert str(stack) == f"1 scans_slider [{value1}, {value3}] 1 frames_slider [{value2}, {value4}] 2 global [None, 'scans_slider', 'frames_slider']", (
            "The update of the second option with a new value doesn't work."
        )
        stack.update("scans_slider", value5)
        assert str(stack) == f"2 scans_slider [{value1}, {value3}, {value5}] 1 frames_slider [{value2}, {value4}] 3 global [None, 'scans_slider', 'frames_slider', 'scans_slider']", (
            "The update of the first option with a new value doesn't work."
        )
        stack.undo()
        assert str(stack) == f"1 scans_slider [{value1}, {value3}, {value5}] 1 frames_slider [{value2}, {value4}] 2 global [None, 'scans_slider', 'frames_slider', 'scans_slider']", (
            "Undo function doesn't work correctly."
        )
        stack.redo()
        assert str(stack) == f"2 scans_slider [{value1}, {value3}, {value5}] 1 frames_slider [{value2}, {value4}] 3 global [None, 'scans_slider', 'frames_slider', 'scans_slider']", (
            "Redo function doesn't work correctly."
        )
        stack.undo()
        stack.update("frames_slider", value6)
        assert str(stack) == f"1 scans_slider [{value1}, {value3}] 2 frames_slider [{value2}, {value4}, {value6}] 3 global [None, 'scans_slider', 'frames_slider', 'frames_slider']", (
            "Update of the values doesn't work after Undo function."
        )


if __name__ == "__main__":
    unittest.main()



# builtin
import os
import dataclasses

# external
import alphatims.bruker
import numpy as np

# local
import alphatims.compiling


@alphatims.compiling.njit_class
@dataclasses.dataclass(kw_only=True, frozen=True)
class DiaData:

    file_name: str = dataclasses.field(repr=False)
    sample_name: str = dataclasses.field(init=False)

    def __init__(
        self,
        *,
        dia_data: alphatims.bruker.TimsTOF,
    ):
        file_name = self.parse_file_name(dia_data.bruker_d_folder_name)
        object.__setattr__(self, "cycle", dia_data._cycle)
        object.__setattr__(self, "im_values", dia_data._mobility_values)
        object.__setattr__(self, "rt_values", dia_data._rt_values)
        object.__setattr__(self, "mz_values", dia_data._mz_values)
        object.__setattr__(self, "frames", dia_data._frames)
        object.__setattr__(self, "fragment_frames", dia_data._fragment_frames)
        object.__setattr__(self, "zeroth_frame", dia_data._zeroth_frame)
        object.__setattr__(self, "tof_indptr", dia_data._push_indptr)
        object.__setattr__(self, "tof_indices", dia_data._tof_indices)
        object.__setattr__(
            self,
            "intensity_values",
            dia_data._intensity_values.astype(np.float32)
        )
        object.__setattr__(self, "as_dataframe", dia_data.as_dataframe)
        object.__setattr__(self, "bin_intensities", dia_data.bin_intensities)
        object.__setattr__(self, "_dia_data", dia_data)

        root_directory = os.path.basename(file_name)
        sample_name = '.'.join(
            os.path.basename(file_name).split('.')[:-1]
        )
        object.__setattr__(self, "file_name", file_name)
        object.__setattr__(self, "sample_name", sample_name)
        object.__setattr__(self, "directory", root_directory)

    def __getitem__(self, keys):
        return self._dia_data[keys]

    @staticmethod
    def parse_file_name(file_name):
        file_name = file_name.strip()
        while file_name[-1] in ["'", '"']:
            file_name = file_name[:-1]
        while file_name[0] in ["'", '"']:
            file_name = file_name[1:]
        if not file_name.endswith(".d"):
            raise ValueError(
                f"File {file_name} is not a valid Bruker .d file."
            )
        return file_name

    def __len__(self):
        return self.size



#!python


# builtin
import contextlib
import os
import logging
import time
import copy
import json

# external
import click

# local
import alphatims
import alphatims.utils

with open(
    os.path.join(alphatims.utils.LIB_PATH, "interface_parameters.json"),
    "r"
) as in_file:
    INTERFACE_PARAMETERS = json.load(in_file)


@contextlib.contextmanager
def parse_cli_settings(command_name: str, **kwargs):
    """A context manager that parses and logs CLI settings.

    Parameters
    ----------
    command_name : str
        The name of the command that utilizes these CLI settings.
    **kwargs
        All values that need to be logged.
        Values (if included) that explicitly will be parsed are:
            output_folder
            threads
            disable_log_stream
            log_file
            parameter_file
            export_parameters

    Returns
    -------
    : dict
        A dictionary with parsed parameters.
    """
    try:
        start_time = time.time()
        kwargs = {key: arg for key, arg in kwargs.items() if arg is not None}
        if "output_folder" in kwargs:
            if kwargs["output_folder"] is not None:
                kwargs["output_folder"] = os.path.abspath(
                    kwargs["output_folder"]
                )
                if not os.path.exists(kwargs["output_folder"]):
                    os.makedirs(kwargs["output_folder"])
                if "log_file" not in kwargs:
                    kwargs["log_file"] = kwargs["output_folder"]
                elif kwargs["log_file"] == "":
                    kwargs["log_file"] = kwargs["output_folder"]
        if ("parameter_file" in kwargs):
            kwargs["parameter_file"] = os.path.abspath(
                kwargs["parameter_file"]
            )
            parameters = alphatims.utils.load_parameters(
                kwargs["parameter_file"]
            )
            kwargs.update(parameters)
        if "threads" not in kwargs:
            kwargs["threads"] = INTERFACE_PARAMETERS[
                "threads"
            ]["default"]
        kwargs["threads"] = alphatims.utils.set_threads(
            kwargs["threads"]
        )
        if "log_file" not in kwargs:
            kwargs["log_file"] = INTERFACE_PARAMETERS[
                "log_file"
            ]["default"]
        if "disable_log_stream" not in kwargs:
            kwargs[
                "disable_log_stream"
            ] = INTERFACE_PARAMETERS[
                "disable_log_stream"
            ]["default"]
        kwargs["log_file"] = alphatims.utils.set_logger(
            log_file_name=kwargs["log_file"],
            stream=not kwargs["disable_log_stream"],
        )
        alphatims.utils.show_platform_info()
        alphatims.utils.show_python_info()
        alphatims.utils.check_github_version()
        logging.info(
            f"Running CLI command `alphatims "
            f"{command_name}` with parameters:"
        )
        max_len = max(len(key) + 1 for key in kwargs)
        for key, value in sorted(kwargs.items()):
            # if (isinstance(value, tuple)) and (len(value) > 1):
            #     key += "s"
            logging.info(f"{key:<{max_len}} - {value}")
        logging.info("")
        if "export_parameters" in kwargs:
            if kwargs["export_parameters"] is not None:
                kwargs["export_parameters"] = os.path.abspath(
                    kwargs["export_parameters"]
                )
                alphatims.utils.save_parameters(
                    kwargs["export_parameters"],
                    kwargs
                )
        yield kwargs
    except Exception:
        logging.exception("Something went wrong, execution incomplete!")
    else:
        logging.info(
            f"Analysis done in {time.time() - start_time:.2f} seconds."
        )
    finally:
        alphatims.utils.set_logger(log_file_name=None)


def cli_option(
    parameter_name: str,
    as_argument: bool = False,
    **kwargs
):
    """A wrapper for click.options and click.arguments using local defaults.

    Parameters
    ----------
    parameter_name : str
        The name of the parameter or argument.
        It's default values need to be present in
        lib/interface_parameters.json.
    as_argument : bool
        If True, a click.argument is returned.
        If False, a click.option is returned.
        Default is False.
    **kwargs
        Items that overwrite the default values of
        lib/interface_parameters.json.
        These need to be valid items for click.
        A special "type" dict can be used to pass a click.Path or click.Choice,
        that has the following format:
        type = {"name": "path" or "choice", **choice_or_path_kwargs}

    Returns
    -------
    : click.option, click.argument
        A click.option or click.argument decorator.
    """
    parameters = copy.deepcopy(
        INTERFACE_PARAMETERS[parameter_name]
    )
    parameters.update(kwargs)
    if "type" in parameters:
        if parameters["type"] == "int":
            parameters["type"] = int
        elif parameters["type"] == "float":
            parameters["type"] = float
        elif parameters["type"] == "str":
            parameters["type"] = str
        elif isinstance(parameters["type"], dict):
            parameter_type = parameters["type"].pop("name")
            if parameter_type == "path":
                parameters["type"] = click.Path(**parameters["type"])
                if ("default" in parameters) and (parameters["default"]):
                    parameters["default"] = os.path.join(
                        os.path.dirname(__file__),
                        parameters["default"]
                    )
            elif parameter_type == "choice":
                options = parameters["type"].pop("options")
                parameters["type"] = click.Choice(
                    options,
                    **parameters["type"]
                )
    if "default" in parameters:
        if "is_flag" in parameters:
            parameters["show_default"] = False
        else:
            parameters["show_default"] = True
    if not as_argument:
        if "short_name" in parameters:
            short_name = parameters.pop("short_name")
            return click.option(
                f"--{parameter_name}",
                f"-{short_name}",
                **parameters,
            )
        else:
            return click.option(
                f"--{parameter_name}",
                **parameters,
            )
    else:
        if "required" in parameters:
            required = parameters.pop("required")
        else:
            required = True
        if "nargs" in parameters:
            return click.argument(
                parameter_name,
                type=click.Path(exists=True),
                nargs=parameters["nargs"],
                required=required,
            )
        else:
            return click.argument(
                parameter_name,
                type=parameters["type"],
                required=required,
            )


@click.group(
    context_settings=dict(
        help_option_names=['-h', '--help'],
    ),
    invoke_without_command=True
)
@click.pass_context
@click.version_option(alphatims.__version__, "-v", "--version")
def run(ctx, **kwargs):
    name = f"AlphaTims {alphatims.__version__}"
    click.echo("*" * (len(name) + 4))
    click.echo(f"* {name} *")
    click.echo("*" * (len(name) + 4))
    if ctx.invoked_subcommand is None:
        click.echo(run.get_help(ctx))


@run.command("gui", help="Start graphical user interface.")
@cli_option("port")
@cli_option("bruker_raw_data", required=False, as_argument=True)
def gui(port, bruker_raw_data):
    with parse_cli_settings("gui"):
        logging.info("Loading GUI..")
        import alphatims.gui
        alphatims.gui.run(port, bruker_raw_data)


@run.group("export", help="Export information.")
def export(**kwargs):
    pass


@run.group(
    "detect",
    help="Detect structures within the data (NotImplemented yet)."
)
def detect(**kwargs):
    pass


@export.command(
    "hdf",
    help="Export BRUKER_RAW_DATA as hdf file.",
    no_args_is_help=True,
)
@cli_option("bruker_raw_data", as_argument=True)
@cli_option("disable_overwrite")
@cli_option("enable_compression")
@cli_option("output_folder")
@cli_option("log_file")
@cli_option("threads")
@cli_option("disable_log_stream")
@cli_option("parameter_file")
@cli_option("export_parameters")
def export_hdf(**kwargs):
    with parse_cli_settings("export hdf", **kwargs) as parameters:
        import alphatims.bruker
        data = alphatims.bruker.TimsTOF(
            parameters["bruker_raw_data"],
            use_hdf_if_available=False
        )
        if "output_folder" not in parameters:
            directory = data.directory
        else:
            directory = parameters["output_folder"]
        data.save_as_hdf(
            overwrite=not parameters["disable_overwrite"],
            directory=directory,
            file_name=f"{data.sample_name}.hdf",
            compress=parameters["enable_compression"],
        )


@export.command(
    "mgf",
    help="Export BRUKER_RAW_DATA as (profile) mgf file.",
    no_args_is_help=True,
)
@cli_option("bruker_raw_data", as_argument=True)
@cli_option("keep_n_most_abundant_peaks")
@cli_option("centroiding_window")
@cli_option("disable_overwrite")
@cli_option("output_folder")
@cli_option("log_file")
@cli_option("threads")
@cli_option("disable_log_stream")
@cli_option("parameter_file")
@cli_option("export_parameters")
def export_mgf(**kwargs):
    with parse_cli_settings("export mgf", **kwargs) as parameters:
        import alphatims.bruker
        data = alphatims.bruker.TimsTOF(parameters["bruker_raw_data"])
        if "output_folder" not in parameters:
            directory = data.directory
        else:
            directory = parameters["output_folder"]
        data.save_as_spectra(
            overwrite=not parameters["disable_overwrite"],
            directory=directory,
            file_name=f"{data.sample_name}.mgf",
            centroiding_window=parameters["centroiding_window"],
            keep_n_most_abundant_peaks=parameters["keep_n_most_abundant_peaks"],
            mgf=True,
        )


@export.command(
    "spectra",
    help="Export BRUKER_RAW_DATA as (profile) spectra file.",
    no_args_is_help=True,
)
@cli_option("bruker_raw_data", as_argument=True)
@cli_option("keep_n_most_abundant_peaks")
@cli_option("centroiding_window")
@cli_option("disable_overwrite")
@cli_option("output_folder")
@cli_option("log_file")
@cli_option("threads")
@cli_option("disable_log_stream")
@cli_option("parameter_file")
@cli_option("export_parameters")
def export_spectra(**kwargs):
    with parse_cli_settings("export spectra", **kwargs) as parameters:
        import alphatims.bruker
        data = alphatims.bruker.TimsTOF(parameters["bruker_raw_data"])
        if "output_folder" not in parameters:
            directory = data.directory
        else:
            directory = parameters["output_folder"]
        data.save_as_spectra(
            overwrite=not parameters["disable_overwrite"],
            directory=directory,
            file_name=f"{data.sample_name}.spectra.hdf",
            centroiding_window=parameters["centroiding_window"],
            keep_n_most_abundant_peaks=parameters["keep_n_most_abundant_peaks"],
            mgf=False,
        )


@export.command(
    "selection",
    help="Load a BRUKER_RAW_DATA and select a data slice for export.",
    no_args_is_help=True,
)
@cli_option("bruker_raw_data", as_argument=True)
@cli_option("ion_type")
@cli_option("rt_bounds")
@cli_option("mobility_bounds")
@cli_option("precursor_bounds")
@cli_option("quad_mz_bounds")
@cli_option("tof_mz_bounds")
@cli_option("intensity_bounds")
@cli_option("format")
@cli_option("x_axis")
@cli_option("y_axis")
@cli_option("output_folder")
@cli_option("log_file")
@cli_option("threads")
@cli_option("disable_log_stream")
@cli_option("parameter_file")
@cli_option("export_parameters")
def export_selection(**kwargs):
    with parse_cli_settings("export selection", **kwargs) as parameters:
        import numpy as np
        import alphatims.bruker
        frame_values = np.empty(
            shape=(len(parameters["rt_bounds"]), 3),
            dtype=np.int64
        )
        scan_values = np.empty(
            shape=(len(parameters["mobility_bounds"]), 3),
            dtype=np.int64
        )
        tof_values = np.empty(
            shape=(len(parameters["tof_mz_bounds"]), 3),
            dtype=np.int64
        )
        intensity_values = np.empty(
            shape=(len(parameters["intensity_bounds"]), 2),
            dtype=np.float64
        )
        if "precursors" in parameters["ion_type"]:
            quad_values = np.array([[-1, 0]])
            precursor_values = np.array([[0, 1, 1]])
        else:
            quad_values = np.empty(
                shape=(len(parameters["quad_mz_bounds"]), 2),
                dtype=np.float64
            )
            precursor_values = np.empty(
                shape=(len(parameters["precursor_bounds"]), 3),
                dtype=np.int64
            )
        logging.info("Loading raw data.")
        data = alphatims.bruker.TimsTOF(parameters["bruker_raw_data"])
        logging.info("Filtering datapoints.")
        for i, rt_bounds in enumerate(parameters["rt_bounds"]):
            if (rt_bounds[0] is not None):
                if (rt_bounds[0] < 0):
                    rt_bounds = (
                        int(-rt_bounds[0]),
                        int(-rt_bounds[1]),
                    )
            frame_values[i] = alphatims.bruker.convert_slice_key_to_int_array(
                data, slice(*rt_bounds), "frame_indices"
            )[0]
        for i, mobility_bounds in enumerate(parameters["mobility_bounds"]):
            if (mobility_bounds[0] is not None):
                if (mobility_bounds[0] < 0):
                    mobility_bounds = (
                        int(-mobility_bounds[0]),
                        int(-mobility_bounds[1]),
                    )
            scan_values[i] = alphatims.bruker.convert_slice_key_to_int_array(
                data, slice(*mobility_bounds), "scan_indices"
            )[0]
        for i, tof_mz_bounds in enumerate(parameters["tof_mz_bounds"]):
            if (tof_mz_bounds[0] is not None):
                if (tof_mz_bounds[0] < 0):
                    tof_mz_bounds = (
                        int(-tof_mz_bounds[0]),
                        int(-tof_mz_bounds[1]),
                    )
            tof_values = alphatims.bruker.convert_slice_key_to_int_array(
                data, slice(*tof_mz_bounds), "tof_indices"
            )
        for i, intensity_bounds in enumerate(parameters["intensity_bounds"]):
            intensity_values[i] = alphatims.bruker.convert_slice_key_to_float_array(
                slice(*intensity_bounds)
            )
        if "fragments" in parameters["ion_type"]:
            for i, quad_mz_bounds in enumerate(parameters["quad_mz_bounds"]):
                quad_values[i] = alphatims.bruker.convert_slice_key_to_float_array(
                    slice(*quad_mz_bounds)
                )
            for i, precursor_bounds in enumerate(parameters["precursor_bounds"]):
                precursor_values[i] = alphatims.bruker.convert_slice_key_to_int_array(
                    data,
                    slice(*precursor_bounds),
                    "precursor_indices"
                )
                if precursor_values[i, 0] < 1:
                    precursor_values[i, 0] = 1
        strike_indices = alphatims.bruker.filter_indices(
            frame_slices=frame_values,
            scan_slices=scan_values,
            precursor_slices=precursor_values,
            tof_slices=tof_values,
            quad_slices=quad_values,
            intensity_slices=intensity_values,
            frame_max_index=data.frame_max_index,
            scan_max_index=data.scan_max_index,
            push_indptr=data.push_indptr,
            precursor_indices=data.precursor_indices,
            quad_mz_values=data.quad_mz_values,
            quad_indptr=data.quad_indptr,
            tof_indices=data.tof_indices,
            intensities=data.intensity_values
        )
        df = data.as_dataframe(strike_indices)
        logging.info(f"Selected {len(df)} datapoints.")
        if "output_folder" not in parameters:
            output_folder = data.directory
        else:
            output_folder = parameters["output_folder"]
        output_file_name_base = os.path.join(
            output_folder,
            f"{data.sample_name}_selection"
        )
        if "csv" in parameters['format']:
            logging.info(f"Exporting results to {output_file_name_base}.csv")
            df.to_csv(f"{output_file_name_base}.csv", index=False)
        if ("html" in parameters['format']) or ("png" in parameters['format']):
            import alphatims.plotting
            import holoviews as hv
            labels = {
                "tof_mz": "mz",
                "rt": "rt",
                "mobility": "mobility",
                "intensity": "intensity",
            }
            x_axis = labels[parameters["x_axis"]]
            y_axis = labels[parameters["y_axis"]]
            if y_axis == "intensity":
                plot = alphatims.plotting.line_plot(
                    data,
                    strike_indices,
                    x_axis,
                    data.sample_name,
                )
            else:
                plot = alphatims.plotting.heatmap(
                    df,
                    x_axis,
                    y_axis,
                    data.sample_name,
                )
            if "html" in parameters['format']:
                logging.info(
                    f"Exporting results to {output_file_name_base}.html"
                )
                hv.save(plot, f"{output_file_name_base}.html")
            if "png" in parameters['format']:
                logging.info(
                    f"Exporting results to {output_file_name_base}.png"
                )
                hv.save(plot, f"{output_file_name_base}.png", fmt="png")


@detect.command(
    "ions",
    help="Detect ions (NotImplemented yet).",
    no_args_is_help=True,
)
@cli_option("bruker_raw_data", as_argument=True)
@cli_option("output_folder")
@cli_option("log_file")
@cli_option("threads")
@cli_option("disable_log_stream")
@cli_option("parameter_file")
@cli_option("export_parameters")
def detect_ions(**kwargs):
    with parse_cli_settings("detect ions", **kwargs) as parameters:
        raise NotImplementedError



#!python
"""This module allows to create temporary mmapped arrays."""

# builtin
import atexit
import logging
import os
import shutil
import tempfile

# external
import mmap
import numpy as np

# local
import alphatims.utils


def make_temp_dir(prefix: str = "temp_mmap_") -> tuple:
    """Make a temporary directory.

    Parameters
    ----------
    prefix : str
        The prefix for the temporary directory.

    Returns
    -------
    tuple
        The temporary directory and its name.
    """
    _temp_dir = tempfile.TemporaryDirectory(prefix=prefix)
    temp_dir_name = _temp_dir.name
    return _temp_dir, temp_dir_name


_TEMP_DIR, TEMP_DIR_NAME = make_temp_dir()
ARRAYS = {}
CLOSED = False
ALLOW_NDARRAY_SUBCLASS = False

def empty(shape: tuple, dtype: np.dtype) -> np.ndarray:
    """Create a writable temporary mmapped array.

    Parameters
    ----------
    shape : tuple
        A tuple with the shape of the array.
    dtype : np.dtype
        The np.dtype of the array.

    Returns
    -------
    np.ndarray
        A writable temporary mmapped array.
    """
    element_count = np.prod(shape, dtype=np.int64)
    if element_count <= 0:
        raise ValueError(
            f"Shape {shape} has an invalid element count of {element_count}"
        )
    *other, free_space = shutil.disk_usage(TEMP_DIR_NAME)
    required_space = element_count * np.dtype(dtype).itemsize
    if free_space < required_space:
        raise IOError(
            f"Cannot create array of size {required_space} "
            f"(available space on {TEMP_DIR_NAME} is {free_space})."
        )
    temp_file_name = os.path.join(
        TEMP_DIR_NAME,
        f"temp_mmap_{np.random.randint(2**31)}{np.random.randint(2**31)}.bin"
    )
    with open(temp_file_name, "wb") as binfile:
        binfile.seek(required_space - 1)
        binfile.write(b"0")
    return read(temp_file_name, element_count, shape, dtype)


def read(
    temp_file_name: str,
    element_count: int,
    shape: tuple,
    dtype: np.dtype,
) -> np.ndarray:
    """Short summary.

    Parameters
    ----------
    temp_file_name : str
        The file name of the temp mmap data.
    element_count : int
        The number of elements to read.
    shape : tuple
        A tuple with the shape of the array.
    dtype : np.dtype
        The np.dtype of the array.

    Returns
    -------
    np.ndarray
        A writable temporary mmapped array.
    """
    with open(temp_file_name, "r+b") as bin_file:
        _mmap_obj = mmap.mmap(
            bin_file.fileno(),
            0,
            access=mmap.ACCESS_WRITE
        )
        _array = np.frombuffer(
            _mmap_obj,
            dtype=dtype,
            count=element_count,
        ).reshape(shape)
        if ALLOW_NDARRAY_SUBCLASS:
            _array = TempMMapArray(_array, temp_file_name)
        ARRAYS[temp_file_name] = (_array, _mmap_obj)
        return _array


def zeros(shape: tuple, dtype: np.dtype) -> np.ndarray:
    """Create a writable temporary mmapped array filled with zeros.

    Parameters
    ----------
    shape : tuple
        A tuple with the shape of the array.
    dtype : type
        The np.dtype of the array.

    Returns
    -------
    type
        A writable temporary mmapped array filled with zeros.
    """
    _array = empty(shape, dtype)
    _array[:] = 0
    return _array


def ones(shape: tuple, dtype: np.dtype) -> np.ndarray:
    """Create a writable temporary mmapped array filled with ones.

    Parameters
    ----------
    shape : tuple
        A tuple with the shape of the array.
    dtype : type
        The np.dtype of the array.

    Returns
    -------
    type
        A writable temporary mmapped array filled with ones.
    """
    _array = empty(shape, dtype)
    _array[:] = 1
    return _array


def arange(
    start: int,
    stop: int,
    step: int = 1,
    dtype: np.dtype = np.uint64
) -> np.ndarray:
    """Create a writable temporary mmapped array filled with as a range.

    Parameters
    ----------
    start : int
        The start of this range.
    stop : int
        The stop of this range.
    step : int
        The step of this range.
    dtype : type
        The np.dtype of the array.

    Returns
    -------
    type
        A writable temporary mmapped array filled as a range.
    """
    _array = empty((stop - start) // step, dtype)
    fill_array(_array, start, stop, step)
    return _array


@alphatims.utils.njit(nogil=True)
def fill_array(array, start, stop, step):
    for index, value in enumerate(range(start, stop, step)):
        array[index] = value


def clone(array: np.ndarray) -> np.ndarray:
    """Create a writable temporary mmapped array copy.

    Parameters
    ----------
    array : np.ndarray
        An array that needs to be copied.

    Returns
    -------
    np.ndarray
        A writable temporary mmapped array that is a copy of array.

    Raises
    ------
    ExceptionName
        Why the exception is raised.

    """
    _array = empty(array.shape, array.dtype)
    _array[:] = array
    return _array


@atexit.register
def atexit_clear() -> str:
    """Reset the temporary folder containing temp mmapped arrays.

    WARNING: All existing temp mmapp arrays will become unusable!

    Returns
    -------
    str
        The name of the new temporary folder.
    """
    global CLOSED
    CLOSED = True
    reset()


def reset() -> str:
    """Reset the temporary folder containing temp mmapped arrays.

    WARNING: All existing temp mmapp arrays will become unusable!

    Returns
    -------
    str
        The name of the new temporary folder.
    """
    global _TEMP_DIR
    global TEMP_DIR_NAME
    global ARRAYS
    global CLOSED
    if not CLOSED:
        logging.warning(
            f"Folder {TEMP_DIR_NAME} with temp mmap arrays is being deleted. "
            "All existing temp mmapp arrays will be unusable!"
        )
    import gc
    for _array_name in list(ARRAYS.keys()):
        _array_tuple = ARRAYS.pop(_array_name)
        _array, _memmap = _array_tuple
        del _array_tuple
        _array.data.release()
        del _array
        gc.collect()
        _memmap.close()
        del _memmap
    del _TEMP_DIR
    _TEMP_DIR, TEMP_DIR_NAME = make_temp_dir()
    ARRAYS = {}
    
    logging.warning(
    f"WARNING: Temp mmap arrays were written to {TEMP_DIR_NAME}. "
    "Cleanup of this folder is OS dependant, "
    "and might need to be triggered manually! "
    f"Current space: {shutil.disk_usage(TEMP_DIR_NAME)[-1]:,}"
    )
    
    return TEMP_DIR_NAME


class TempMMapArray(np.ndarray):
    """A np.ndarray with path attribute for a temporary mmapped buffer."""

    def __new__(cls, input_array, _path):
        obj = np.asarray(input_array).view(cls)
        obj._path = _path
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._path = getattr(obj, '_path', None)

    @property
    def path(self):
        return self._path



#!python

# builtin
import setuptools
import re
# local
import alphatims as package2install


with open("README.md", "r") as readme_file:
    LONG_DESCRIPTION = readme_file.read()

extra_requirements = {}
for extra, requirement_file_name in package2install.__requirements__.items():
    with open(requirement_file_name) as requirements_file:
        if extra != "":
            extra_stable = f"{extra}-stable"
        else:
            extra_stable = "stable"
        extra_requirements[extra_stable] = []
        extra_requirements[extra] = []
        for line in requirements_file:
            extra_requirements[extra_stable].append(line)
            # conditional req like: "pywin32; sys_platform=='win32'"
            line, *conditions = line.split(';')
            requirement, *comparison = re.split("[><=~!]", line)
            requirement == requirement.strip()
            requirement = ";".join([requirement] + conditions)
            extra_requirements[extra].append(requirement)

requirements = extra_requirements.pop("")

setuptools.setup(
    name=package2install.__project__,
    version=package2install.__version__,
    license=package2install.__license__,
    description=package2install.__description__,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author=package2install.__author__,
    author_email=package2install.__author_email__,
    url=package2install.__github__,
    project_urls=package2install.__urls__,
    keywords=package2install.__keywords__,
    classifiers=package2install.__classifiers__,
    packages=[package2install.__project__],
    include_package_data=True,
    entry_points={
        "console_scripts": package2install.__console_scripts__,
    },
    install_requires=requirements,
    extras_require=extra_requirements,
    python_requires=package2install.__python_version__,
)



# builtin
import ast
import textwrap
import inspect
import types
import weakref

# external
import numba
import pandas as pd
import numpy as np


def precompile_njit_functions_from_object(object_):
    if not is_regular_object_with_dict(object_):
        return
    create_njit_module_for_object(object_)
    for func in iterate_over_callables_from_object(object_):
        try:
            tree = get_source_tree_from_callable(func)
        except OSError:
            pass
        else:
            if tree_has_decorator_containing_name(tree, "njit"):
                add_njit_function_to_object_njit_module(object_, func, tree)
                overwrite_object_function_with_njit_function(object_, func)


def is_regular_object_with_dict(object_):
    if not hasattr(object_, "__dict__"):
        return False
    if isinstance(object_, numba.core.registry.CPUDispatcher):
        return False
    return True


def create_njit_module_for_object(object_):
    module = create_module(object_)
    module.__dict__.update(inspect.getmodule(object_).__dict__)
    module.self = module
    object.__setattr__(object_, "__njit__", module)


def create_module(x):
    if has_njit_module(x):
        return x.__njit__
    if is_pandas_dataframe(x):
        return create_module_from_dataframe(x)
    elif is_regular_object_with_dict(x):
        return create_module_from_object(x)
    else:
        return x


def is_module(x):
    return hasattr(x, "__njit__")


def has_njit_module(x):
    return isinstance(x, types.ModuleType)


def is_pandas_dataframe(x):
    return isinstance(x, pd.DataFrame)


def create_module_from_dataframe(df):
    module = types.ModuleType(f"{df.__class__}_{id(df)}")
    for column in df.columns:
        module.__dict__[column] = np.array(df[column].values, copy=False)
    for key, value in df.__dict__.items():
        if not key.startswith("_"):
            module.__dict__[key] = create_module(value)
    return module


def create_module_from_object(object_):
    module = types.ModuleType(f"{object_.__class__}_{id(object_)}")
    for key, value in object_.__dict__.items():
        if not key.startswith("__"):
            module.__dict__[key] = create_module(value)
    return module


def iterate_over_callables_from_object(object_):
    for key in list(dir(object_)):
        if key.startswith("__"):
            continue
        potential_func = eval(f"object_.{key}")
        if callable(potential_func):
            yield potential_func


def get_source_tree_from_callable(func):
    src = inspect.getsource(func)
    src = textwrap.dedent(src)
    return ast.parse(src)


def tree_has_decorator_containing_name(tree, name):
    for decorator in tree.body[0].decorator_list:
        if decorator_contains(decorator, name):
            return True
    return False


def decorator_contains(decorator, name):
    # TODO: Should be properly parsed!
    if name in ast.unparse(decorator):
        return True
    return False


def add_njit_function_to_object_njit_module(object_, func, tree):
    src = create_src_without_self_and_decorators_from_function_tree(tree)
    exec(src, object_.__njit__.__dict__)
    nogil = tree_has_decorator_containing_name(tree, "nogil")
    func_ = numba.njit(nogil=nogil)(object_.__njit__.__dict__[func.__name__])
    src = f"object.__setattr__(object_.__njit__, '{func.__name__}', func_)"
    exec(src)


def create_src_without_self_and_decorators_from_function_tree(tree):
    # TODO: removes the nogil decorator as well!
    origonal_decorators = tree.body[0].decorator_list
    origonal_args = tree.body[0].args.args
    tree.body[0].decorator_list = []
    tree.body[0].args.args = tree.body[0].args.args[1:]
    src = ast.unparse(tree)
    tree.body[0].decorator_list = origonal_decorators
    tree.body[0].args.args = origonal_args
    return src


def overwrite_object_function_with_njit_function(object_, func):
    src = f"object.__setattr__(object_, '{func.__name__}', object_.__njit__.{func.__name__})"
    exec(src)


def njit(*args, **kwargs):
    return numba.njit(*args, **kwargs)


# def njit_class(_func=None, dataclass=False, **decorator_kwargs):
#     import functools
#     def wrapper(_func):
#         @functools.wraps(_func)
#         def inner_func(*func_args, **func_kwargs):
#             _object = _func(*func_args,  **func_kwargs)
#             precompile_njit_functions_from_object(_object)
#             return _object
#         return inner_func
#     if _func is None:
#         return wrapper
#     else:
#         return wrapper(_func)


def njit_class(_cls=None, njit=True):
    def wrapper(_cls):
        def __blank_post_init__(self):
            if hasattr(super(self.__class__.__mro__[self.__class_index__], self), "__post_init__"):
                super(self.__class__.__mro__[self.__class_index__], self).__post_init__()
        def __post_init__(self):
            if not hasattr(self, "__class_index__"):
                object.__setattr__(self, "__class_index__", -1)
            object.__setattr__(self, "__class_index__", self.__class_index__ + 1)
            self.__class__.__mro__[self.__class_index__].__original_post_init__(self)
            object.__setattr__(self, "__class_index__", self.__class_index__ - 1)
            if self.__class_index__ == -1:
                if njit and not hasattr(self, "__njit__"):
                    self.precompile_njit_functions_from_object()
                del self.__dict__["__class_index__"]
        _cls.precompile_njit_functions_from_object = precompile_njit_functions_from_object
        for _super_cls in _cls.__mro__[:-1][::-1]:
            has_post_init = ("__post_init__" in _super_cls.__dict__)
            if not has_post_init:
                _super_cls.__post_init__ = __blank_post_init__
            has_original_post_init = ("__original_post_init__" in _super_cls.__dict__)
            if not has_original_post_init:
                _super_cls.__original_post_init__ = _super_cls.__post_init__
                _super_cls.__post_init__ = __post_init__
        return _cls
    if _cls is None:
        return wrapper
    else:
        return wrapper(_cls)



# def njit_class(_cls=None, dataclass=False, njit=True, hdf=True, **decorator_kwargs):
#     import functools
#     def wrapper(_cls):
#         def __blank_post_init__(self):
#             if hasattr(super(self.__class__.__mro__[self.__class_index__], self), "__post_init__"):
#                 super(self.__class__.__mro__[self.__class_index__], self).__post_init__()
#         def __post_init__(self):
#             if not hasattr(self, "__class_index__"):
#                 object.__setattr__(self, "__class_index__", -1)
#             object.__setattr__(self, "__class_index__", self.__class_index__ + 1)
#             self.__class__.__mro__[self.__class_index__].__original_post_init__(self)
#             object.__setattr__(self, "__class_index__", self.__class_index__ - 1)
#             if self.__class_index__ == -1:
#                 if njit and not hasattr(self, "__njit__"):
#                     tdf2ms2.utils.compiling.precompile_njit_functions_from_object(self)
#                 if hdf and not hasattr(self, "__hdf__"):
#                     save_to_hdf(self)
#                 del self.__dict__["__class_index__"]
#         for _super_cls in _cls.__mro__[:-1][::-1]:
#             has_post_init = ("__post_init__" in _super_cls.__dict__)
#             if not has_post_init:
#                 _super_cls.__post_init__ = __blank_post_init__
#             has_original_post_init = ("__original_post_init__" in _super_cls.__dict__)
#             if not has_original_post_init:
#                 _super_cls.__original_post_init__ = _super_cls.__post_init__
#                 _super_cls.__post_init__ = __post_init__
#         return dataclasses.dataclass(**decorator_kwargs)(_cls)
#     if _cls is None:
#         return wrapper
#     else:
#         return wrapper(_cls)


# def save_to_hdf(self, file_name=None):
#     if file_name is None:
#         file_name = f"sandbox/{hash(self)}.hdf"
#     object.__setattr__(self, "__hdf__", file_name)
#     print(f"Saving object results to {file_name}.")
#     import tdf2ms2.utils.hdf
#     import inspect
#     hdf = tdf2ms2.utils.hdf.HDF_File(
#         file_name,
#         read_only=False,
#         truncate=True,
#     )
#     for key, value in self.__dict__.items():
#         if not callable(value):
#             if hasattr(value, "__hdf__"):
#                 hdf.__setattr__(key, value.__hdf__)
#             elif inspect.ismodule(value):
#                 continue
#             else:
#                 hdf.__setattr__(key, value)



#!python -m unittest tests.test_bruker
"""This module provides unit tests for alphatims.utils."""

# builtin
import unittest
import logging
import os

# external
import numpy as np

# local
import alphatims.utils
import alphatims.bruker
alphatims.utils.set_progress_callback(None)


class TestSlicing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(alphatims.utils.DEMO_FILE_NAME):
            logging.info("Downloading sample...")
            import urllib.request
            import urllib.error
            import zipfile
            import io
            with urllib.request.urlopen(
                alphatims.utils.DEMO_FILE_NAME_GITHUB
            ) as sample_file:
                sample_byte_stream = io.BytesIO(sample_file.read())
                with zipfile.ZipFile(sample_byte_stream, 'r') as zip_ref:
                    zip_ref.extractall(
                        os.path.dirname(alphatims.utils.DEMO_FILE_NAME)
                    )
        if os.path.exists(alphatims.utils.DEMO_FILE_NAME):
            try:
                cls.data = alphatims.bruker.TimsTOF(
                    alphatims.utils.DEMO_FILE_NAME
                )
            except:
                assert False, "Test data set is invalid..."
        else:
            assert False, "No test data found..."
            # TODO: fetch from URL?

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_convert_slice_key_to_float_array(self):
        test_cases = [
            (
                slice(100, 300),
                np.array([[100., 300.]])
            ), (
                slice(300),
                np.array([[-np.inf, 300.]])
            ), (
                slice(None, 300),
                np.array([[-np.inf, 300.]])
            ), (
                slice(300, None),
                np.array([[300., np.inf]])
            ), (
                100,
                np.array([[100., 100.]])
            ), (
                100.8,
                np.array([[100.8, 100.8]])
            ), (
                [100, 300],
                np.array([[100., 100.], [300., 300.]])
            ), (
                np.array([[100., 100.], [300., 300.]]),
                np.array([[100., 100.], [300., 300.]])
            ),
        ]
        for key, expected_result in test_cases:
            result = alphatims.bruker.convert_slice_key_to_float_array(
                key
            )
            assert np.array_equal(
                result,
                expected_result
            ), (
                f"{key} was wrongly converted to {result} "
                f"instead of {expected_result}"
            )

    def test_convert_slice_key_to_int_array(self):
        for dimension, (float_low, float_high, max_int) in {
            "frame_indices": [110., 150., self.data.frame_max_index],
            "tof_indices": [500., 500.5, self.data.tof_max_index],
        }.items():
            int_low, int_high = self.data.convert_to_indices(
                [float_low, float_high],
                return_frame_indices=dimension == "frame_indices",
                return_scan_indices=dimension == "scan_indices",
                return_tof_indices=dimension == "tof_indices",
            )
            for key, expected_result in [
                (
                    slice(int_low, int_high),
                    np.array([[int_low, int_high, 1]])
                ), (
                    slice(int_low, int_high, 8),
                    np.array([[int_low, int_high, 8]])
                ), (
                    slice(None, int_high, 8),
                    np.array([[0, int_high, 8]])
                ), (
                    slice(int_low, None, 8),
                    np.array([[int_low, max_int, 8]])
                ), (
                    slice(None, None, 8),
                    np.array([[0, max_int, 8]])
                ), (
                    int_low,
                    np.array([[int_low, int_low + 1, 1]])
                ), (
                    [int_low, int_high],
                    np.array(
                        [
                            [int_low, int_low + 1, 1],
                            [int_high, int_high + 1, 1]
                        ]
                    )
                ), (
                    slice(None, float_high, 8),
                    np.array([[0, int_high, 8]])
                ), (
                    slice(float_low, None, 8),
                    np.array([[int_low, max_int, 8]])
                ), (
                    slice(float_low, float_high, 8),
                    np.array([[int_low, int_high, 8]])
                ), (
                    [float_low, float_high],
                    np.array(
                        [
                            [int_low, int_low + 1, 1],
                            [int_high, int_high + 1, 1]
                        ]
                    )
                ),
            ]:
                result = alphatims.bruker.convert_slice_key_to_int_array(
                    self.data,
                    key,
                    dimension
                )
                assert np.array_equal(
                    result,
                    expected_result
                ), (
                    f"Key '{key}' in dimension '{dimension}' was wrongly "
                    f"converted to {result} instead of "
                    f"{expected_result}"
                )
        dimension, float_low, float_high, max_int = [
            "scan_indices", 1., 1.2, self.data.scan_max_index
        ]
        # NOTE: order high and low is reversed
        int_high, int_low = self.data.convert_to_indices(
            [float_low, float_high],
            return_frame_indices=dimension == "frame_indices",
            return_scan_indices=dimension == "scan_indices",
            return_tof_indices=dimension == "tof_indices",
        )
        for key, expected_result in [
            (
                slice(int_low, int_high),
                np.array([[int_low, int_high, 1]])
            ), (
                slice(int_low, int_high, 8),
                np.array([[int_low, int_high, 8]])
            ), (
                slice(None, int_high, 8),
                np.array([[0, int_high, 8]])
            ), (
                slice(int_low, None, 8),
                np.array([[int_low, max_int, 8]])
            ), (
                slice(None, None, 8),
                np.array([[0, max_int, 8]])
            ), (
                int_low,
                np.array([[int_low, int_low + 1, 1]])
            ), (
                [int_low, int_high],
                np.array(
                    [
                        [int_low, int_low + 1, 1],
                        [int_high, int_high + 1, 1]
                    ]
                )
            ), (
                slice(None, float_high, 8),
                np.array([[int_low, max_int, 8]])
            ), (
                slice(float_low, None, 8),
                np.array([[0, int_high, 8]])
            ), (
                slice(float_low, float_high, 8),
                np.array([[int_low, int_high, 8]])
            ), (
                [float_high, float_low],
                np.array(
                    [
                        [int_low - 1, int_low, 1],
                        [int_high - 1, int_high, 1]
                    ]
                )
            ),
        ]:
            result = alphatims.bruker.convert_slice_key_to_int_array(
                self.data,
                key,
                dimension
            )
            assert np.array_equal(result, expected_result), (
                f"Key '{key}' in dimension '{dimension}' was wrongly "
                f"converted to {result} instead of "
                f"{expected_result}"
            )

    def test_parse_keys(self):
        # TODO: expand testing?
        key = (1,)
        result = alphatims.bruker.parse_keys(self.data, key)
        expected_result = {
            "frame_indices": np.array(
                [[1, 2, 1]],
                dtype=np.int64
            ),
            "scan_indices": np.array(
                [[0, self.data.scan_max_index, 1]],
                dtype=np.int64
            ),
            "tof_indices": np.array(
                [[0, self.data.tof_max_index, 1]],
                dtype=np.int64
            ),
            "precursor_indices": np.array(
                [[0, self.data.precursor_max_index, 1]],
                dtype=np.int64
            ),
            "quad_values": np.array(
                [[-np.inf, np.inf]],
                dtype=np.float64
            ),
            "intensity_values": np.array(
                [[-np.inf, np.inf]],
                dtype=np.float64
            ),
        }
        for dimension in [
            "frame_indices",
            "scan_indices",
            "tof_indices",
            "precursor_indices",
            "quad_values",
            "intensity_values",
        ]:
            assert np.array_equal(
                result[dimension],
                expected_result[dimension]
            ), (
                f"Key '{key}' in dimension '{dimension}' was wrongly "
                f"converted to {result[dimension]} instead of "
                f"{expected_result[dimension]}"
            )

    def test_data_slicing(self):
        df = self.data[1, :700, 0, 500.1:600.5, [100, 200]]
        min_values = df.min()
        assert min_values.frame_indices >= 1
        assert min_values.mz_values >= 500.1
        max_values = df.max()
        assert max_values.frame_indices < 2
        assert max_values.scan_indices < 700
        assert max_values.mz_values < 600.5
        assert len(
            set(np.unique(df.intensity_values)) - set([100, 200])
        ) == 0
        df = self.data[:100., 1.:]
        assert np.min(df.mobility_values) >= 1
        assert np.min(df.rt_values) < 100.
        df = self.data[:100., :1.]
        assert np.min(df.mobility_values) < 1
        assert np.min(df.rt_values) < 100.
        # TEST


if __name__ == "__main__":
    unittest.main()
